@@ -1,65 +1,55 @@
-from __future__ import absolute_import, unicode_literals
-
-from abc import ABCMeta, abstractmethod
-from calendar import timegm
-from datetime import datetime
-
-import jwt
-from six import string_types
-from requests_toolbelt import MultipartEncoder
-
-from arango.exceptions import (
-    ServerConnectionError,
-    JWTAuthError,
-)
-from arango.request import Request
-from arango.response import Response
-
 __all__ = [
-    'Connection',
-    'BasicConnection',
-    'JWTConnection',
-    'JWTSuperuserConnection'
+    "BaseConnection",
+    "BasicConnection",
+    "Connection",
+    "JwtConnection",
+    "JwtSuperuserConnection",
 ]
 
+import sys
+from abc import abstractmethod
+from calendar import timegm
+from datetime import datetime
+from typing import Any, Callable, Optional, Sequence, Union
 
-class Connection(object):
-    """Base connection to specific ArangoDB database.
+import jwt
+from requests import Session
+from requests_toolbelt import MultipartEncoder
 
-    :param hosts: Host URL or list of URLs (coordinators in a cluster).
-    :type hosts: [str]
-    :param host_resolver: Host resolver (used for clusters).
-    :type host_resolver: arango.resolver.HostResolver
-    :param sessions: HTTP session objects per host.
-    :type sessions: [requests.Session]
-    :param db_name: Database name.
-    :type db_name: str
-    :param http_client: User-defined HTTP client.
-    :type http_client: arango.http.HTTPClient
-    """
+from arango.exceptions import JWTAuthError, ServerConnectionError
+from arango.http import HTTPClient
+from arango.request import Request
+from arango.resolver import HostResolver
+from arango.response import Response
+from arango.typings import Fields, Json
 
-    __metaclass__ = ABCMeta
+Connection = Union["BasicConnection", "JwtConnection", "JwtSuperuserConnection"]
 
-    auth_required_header = 'WWW-Authenticate'
 
-    def __init__(self,
-                 hosts,
-                 host_resolver,
-                 sessions,
-                 db_name,
-                 http_client,
-                 serializer,
-                 deserializer):
-        self._url_prefixes = ['{}/_db/{}'.format(h, db_name) for h in hosts]
+class BaseConnection:
+    """Base connection to a specific ArangoDB database."""
+
+    def __init__(
+        self,
+        hosts: Fields,
+        host_resolver: HostResolver,
+        sessions: Sequence[Session],
+        db_name: str,
+        http_client: HTTPClient,
+        serializer: Callable[..., str],
+        deserializer: Callable[[str], Any],
+    ) -> None:
+        self._url_prefixes = [f"{host}/_db/{db_name}" for host in hosts]
         self._host_resolver = host_resolver
         self._sessions = sessions
         self._db_name = db_name
         self._http = http_client
         self._serializer = serializer
         self._deserializer = deserializer
+        self._username: Optional[str] = None
 
     @property
-    def db_name(self):
+    def db_name(self) -> str:
         """Return the database name.
 
         :returns: Database name.
@@ -67,30 +57,39 @@ class Connection(object):
         """
         return self._db_name
 
-    def serialize(self, obj):
-        """Serialize the object and return the string.
+    @property
+    def username(self) -> Optional[str]:
+        """Return the username.
 
-        :param obj: Object to serialize.
-        :type obj: str | bool | int | list | dict
+        :returns: Username.
+        :rtype: str
+        """
+        return self._username
+
+    def serialize(self, obj: Any) -> str:
+        """Serialize the given object.
+
+        :param obj: JSON object to serialize.
+        :type obj: str | bool | int | float | list | dict | None
         :return: Serialized string.
         :rtype: str
         """
         return self._serializer(obj)
 
-    def deserialize(self, string):
+    def deserialize(self, string: str) -> Any:
         """De-serialize the string and return the object.
 
         :param string: String to de-serialize.
         :type string: str
-        :return: De-serialized object.
-        :rtype: str | bool | int | list | dict
+        :return: De-serialized JSON object.
+        :rtype: str | bool | int | float | list | dict | None
         """
         try:
             return self._deserializer(string)
         except (ValueError, TypeError):
             return string
 
-    def prep_response(self, resp, deserialize=True):
+    def prep_response(self, resp: Response, deserialize: bool = True) -> Response:
         """Populate the response with details and return it.
 
         :param deserialize: Deserialize the response body.
@@ -103,8 +102,8 @@ class Connection(object):
         if deserialize:
             resp.body = self.deserialize(resp.raw_body)
             if isinstance(resp.body, dict):
-                resp.error_code = resp.body.get('errorNum')
-                resp.error_message = resp.body.get('errorMessage')
+                resp.error_code = resp.body.get("errorNum")
+                resp.error_message = resp.body.get("errorMessage")
         else:
             resp.body = resp.raw_body
 
@@ -112,7 +111,7 @@ class Connection(object):
         resp.is_success = http_ok and resp.error_code is None
         return resp
 
-    def prep_bulk_err_response(self, parent_response, body):
+    def prep_bulk_err_response(self, parent_response: Response, body: Json) -> Response:
         """Build and return a bulk error response.
 
         :param parent_response: Parent response.
@@ -131,46 +130,42 @@ class Connection(object):
             raw_body=self.serialize(body),
         )
         resp.body = body
-        resp.error_code = body['errorNum']
-        resp.error_message = body['errorMessage']
+        resp.error_code = body["errorNum"]
+        resp.error_message = body["errorMessage"]
         resp.is_success = False
         return resp
 
-    def get_normalized_data(self, request):
+    def normalize_data(self, data: Any) -> Union[str, MultipartEncoder, None]:
         """Normalize request data.
 
-        :param request: HTTP request.
-        :type request: arango.request.Request
+        :param data: Request data.
+        :type data: str | MultipartEncoder | None
         :return: Normalized data.
-        :rtype: str | bool | int | list | dict
+        :rtype: str | MultipartEncoder | None
         """
-        if request.data is None:
-            return request.data
-        elif isinstance(request.data, (string_types, MultipartEncoder)):
-            return request.data
+        if data is None:
+            return None
+        elif isinstance(data, (str, MultipartEncoder)):
+            return data
         else:
-            return self.serialize(request.data)
+            return self.serialize(data)
 
-    def ping(self):
+    def ping(self) -> int:
         """Ping the next host to check if connection is established.
 
         :return: Response status code.
         :rtype: int
         """
-        request = Request(
-            method='get',
-            endpoint='/_api/collection'
-        )
+        request = Request(method="get", endpoint="/_api/collection")
         resp = self.send_request(request)
         if resp.status_code in {401, 403}:
-            raise ServerConnectionError('bad username and/or password')
+            raise ServerConnectionError("bad username and/or password")
         if not resp.is_success:  # pragma: no cover
-            raise ServerConnectionError(
-                resp.error_message or 'bad server response')
+            raise ServerConnectionError(resp.error_message or "bad server response")
         return resp.status_code
 
     @abstractmethod
-    def send_request(self, request):  # pragma: no cover
+    def send_request(self, request: Request) -> Response:  # pragma: no cover
         """Send an HTTP request to ArangoDB server.
 
         :param request: HTTP request.
@@ -181,7 +176,7 @@ class Connection(object):
         raise NotImplementedError
 
 
-class BasicConnection(Connection):
+class BasicConnection(BaseConnection):
     """Connection to specific ArangoDB database using basic authentication.
 
     :param hosts: Host URL or list of URLs (coordinators in a cluster).
@@ -200,38 +195,31 @@ class BasicConnection(Connection):
     :type http_client: arango.http.HTTPClient
     """
 
-    def __init__(self,
-                 hosts,
-                 host_resolver,
-                 sessions,
-                 db_name,
-                 username,
-                 password,
-                 http_client,
-                 serializer,
-                 deserializer):
-        super(BasicConnection, self).__init__(
+    def __init__(
+        self,
+        hosts: Fields,
+        host_resolver: HostResolver,
+        sessions: Sequence[Session],
+        db_name: str,
+        username: str,
+        password: str,
+        http_client: HTTPClient,
+        serializer: Callable[..., str],
+        deserializer: Callable[[str], Any],
+    ) -> None:
+        super().__init__(
             hosts,
             host_resolver,
             sessions,
             db_name,
             http_client,
             serializer,
-            deserializer
+            deserializer,
         )
         self._username = username
         self._auth = (username, password)
 
-    @property
-    def username(self):
-        """Return the username.
-
-        :returns: Username.
-        :rtype: str
-        """
-        return self._username
-
-    def send_request(self, request):
+    def send_request(self, request: Request) -> Response:
         """Send an HTTP request to ArangoDB server.
 
         :param request: HTTP request.
@@ -245,14 +233,14 @@ class BasicConnection(Connection):
             method=request.method,
             url=self._url_prefixes[host_index] + request.endpoint,
             params=request.params,
-            data=self.get_normalized_data(request),
+            data=self.normalize_data(request.data),
             headers=request.headers,
-            auth=self._auth
+            auth=self._auth,
         )
         return self.prep_response(resp, request.deserialize)
 
 
-class JWTConnection(Connection):
+class JwtConnection(BaseConnection):
     """Connection to specific ArangoDB database using JWT authentication.
 
     :param hosts: Host URL or list of URLs (coordinators in a cluster).
@@ -271,35 +259,38 @@ class JWTConnection(Connection):
     :type http_client: arango.http.HTTPClient
     """
 
-    def __init__(self,
-                 hosts,
-                 host_resolver,
-                 sessions,
-                 db_name,
-                 username,
-                 password,
-                 http_client,
-                 serializer,
-                 deserializer):
-        super(JWTConnection, self).__init__(
+    def __init__(
+        self,
+        hosts: Fields,
+        host_resolver: HostResolver,
+        sessions: Sequence[Session],
+        db_name: str,
+        username: str,
+        password: str,
+        http_client: HTTPClient,
+        serializer: Callable[..., str],
+        deserializer: Callable[[str], Any],
+    ) -> None:
+        super().__init__(
             hosts,
             host_resolver,
             sessions,
             db_name,
             http_client,
             serializer,
-            deserializer
+            deserializer,
         )
         self._username = username
         self._password = password
-        self.exp_leeway = 0
 
-        self._auth_header = None
-        self._token = None
-        self._token_exp = None
+        self.exp_leeway: int = 0
+        self._auth_header: Optional[str] = None
+        self._token: Optional[str] = None
+        self._token_exp: int = sys.maxsize
+
         self.refresh_token()
 
-    def send_request(self, request):
+    def send_request(self, request: Request) -> Response:
         """Send an HTTP request to ArangoDB server.
 
         :param request: HTTP request.
@@ -308,14 +299,16 @@ class JWTConnection(Connection):
         :rtype: arango.response.Response
         """
         host_index = self._host_resolver.get_host_index()
-        request.headers['Authorization'] = self._auth_header
+
+        if self._auth_header is not None:
+            request.headers["Authorization"] = self._auth_header
 
         resp = self._http.send_request(
             session=self._sessions[host_index],
             method=request.method,
             url=self._url_prefixes[host_index] + request.endpoint,
             params=request.params,
-            data=self.get_normalized_data(request),
+            data=self.normalize_data(request.data),
             headers=request.headers,
         )
         resp = self.prep_response(resp, request.deserialize)
@@ -329,61 +322,65 @@ class JWTConnection(Connection):
             return resp
 
         self.refresh_token()
-        request.headers['Authorization'] = self._auth_header
+
+        if self._auth_header is not None:
+            request.headers["Authorization"] = self._auth_header
 
         resp = self._http.send_request(
             session=self._sessions[host_index],
             method=request.method,
             url=self._url_prefixes[host_index] + request.endpoint,
             params=request.params,
-            data=self.get_normalized_data(request),
+            data=self.normalize_data(request.data),
             headers=request.headers,
         )
         return self.prep_response(resp, request.deserialize)
 
-    def refresh_token(self):
+    def refresh_token(self) -> None:
         """Get a new JWT token for the current user (cannot be a superuser).
 
         :return: JWT token.
         :rtype: str
         """
         request = Request(
-            method='post',
-            endpoint='/_open/auth',
-            data={
-                'username': self._username,
-                'password': self._password
-            }
+            method="post",
+            endpoint="/_open/auth",
+            data={"username": self._username, "password": self._password},
         )
+
         host_index = self._host_resolver.get_host_index()
+
         resp = self._http.send_request(
             session=self._sessions[host_index],
             method=request.method,
             url=self._url_prefixes[host_index] + request.endpoint,
-            data=self.get_normalized_data(request)
+            data=self.normalize_data(request.data),
         )
         resp = self.prep_response(resp)
+
         if not resp.is_success:
             raise JWTAuthError(resp, request)
 
-        self._token = resp.body['jwt']
+        self._token = resp.body["jwt"]
+        assert self._token is not None
+
         jwt_payload = jwt.decode(
             self._token,
-            issuer='arangodb',
-            algorithms=['HS256'],
+            issuer="arangodb",
+            algorithms=["HS256"],
             options={
-                'require_exp': True,
-                'require_iat': True,
-                'verify_iat': True,
-                'verify_exp': True,
-                'verify_signature': False
+                "require_exp": True,
+                "require_iat": True,
+                "verify_iat": True,
+                "verify_exp": True,
+                "verify_signature": False,
             },
         )
-        self._token_exp = jwt_payload['exp']
-        self._auth_header = 'bearer {}'.format(self._token)
+        self._token_exp = jwt_payload["exp"]
+        self._auth_header = f"bearer {self._token}"
 
 
-class JWTSuperuserConnection(Connection):
+class JwtSuperuserConnection(BaseConnection):
     """Connection to specific ArangoDB database using superuser JWT.
 
     :param hosts: Host URL or list of URLs (coordinators in a cluster).
@@ -400,27 +397,29 @@ class JWTSuperuserConnection(Connection):
     :type superuser_token: str
     """
 
-    def __init__(self,
-                 hosts,
-                 host_resolver,
-                 sessions,
-                 db_name,
-                 http_client,
-                 serializer,
-                 deserializer,
-                 superuser_token):
-        super(JWTSuperuserConnection, self).__init__(
+    def __init__(
+        self,
+        hosts: Fields,
+        host_resolver: HostResolver,
+        sessions: Sequence[Session],
+        db_name: str,
+        http_client: HTTPClient,
+        serializer: Callable[..., str],
+        deserializer: Callable[[str], Any],
+        superuser_token: str,
+    ) -> None:
+        super().__init__(
             hosts,
             host_resolver,
             sessions,
             db_name,
             http_client,
             serializer,
-            deserializer
+            deserializer,
         )
-        self._auth_header = 'bearer {}'.format(superuser_token)
+        self._auth_header = f"bearer {superuser_token}"
 
-    def send_request(self, request):
+    def send_request(self, request: Request) -> Response:
         """Send an HTTP request to ArangoDB server.
 
         :param request: HTTP request.
@@ -429,14 +428,14 @@ class JWTSuperuserConnection(Connection):
         :rtype: arango.response.Response
         """
         host_index = self._host_resolver.get_host_index()
-        request.headers['Authorization'] = self._auth_header
+        request.headers["Authorization"] = self._auth_header
 
         resp = self._http.send_request(
             session=self._sessions[host_index],
             method=request.method,
             url=self._url_prefixes[host_index] + request.endpoint,
             params=request.params,
-            data=self.get_normalized_data(request),
+            data=self.normalize_data(request.data),
             headers=request.headers,
         )
         return self.prep_response(resp, request.deserialize)
